
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

# adapted from: https://gist.github.com/KrzysztofCiba/4579691

##
# @file TypedList.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2012/07/19 08:21:22
# @brief Definition of TypedList class.


class tlist(list):
    """
    .. class:: tlist

    A list-like class holding only objects of specified type(s).
    """

    def __init__(self, iterable=None, types=None):
        iterable = list() if not iterable else iterable

        # make sure it is iterable
        iter(iterable)

        types = types if isinstance(types, tuple) else (types,)
        for item in types:
            if not isinstance(item, type):
                raise TypeError("%s is not a type" % repr(item))

        self._types = types
        for i in iterable:
            self._type_check(i)
        list.__init__(self, iterable)

    def types(self):
        return self._types

    def _type_check(self, val):
        if not self._types:
            return

        if not isinstance(val, self._types):
            raise TypeError(
                "Wrong type %s, this list can hold only instances of %s"
                                                % (type(val), str(self._types)))

    def __iadd__(self, other):
        map(self._type_check, other)
        list.__iadd__(self, other)
        return self

    def __add__(self, other):
        iterable = [item for item in self] + [item for item in other]
        return tlist(iterable, self._types)

    def __radd__(self, other):
        iterable = [item for item in other] + [item for item in self]
        if isinstance(other, tlist):
            return self.__class__(iterable, other.types())
        return tlist(iterable, self._types)

    def __setitem__(self, key, value):
        itervalue = (value,)
        if isinstance(key, slice):
            iter(value)
            itervalue = value
        map(self._type_check, itervalue)
        list.__setitem__(self, key, value)

    def __setslice__(self, i, j, iterable):
        iter(iterable)
        map(self._type_check, iterable)
        list.__setslice__(self, i, j, iterable)

    def append(self, val):
        self._type_check(val)
        list.append(self, val)

    def extend(self, iterable):
        iter(iterable)
        map(self._type_check, iterable)
        list.extend(self, iterable)

    def insert(self, i, val):
        self._type_check(val)
        list.insert(self, i, val)



# encoding: utf8
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from collections import defaultdict

from email import utils
from email.utils import encode_rfc2231
from email.message import tspecials

from spyne import TransportContext, MethodDescriptor, MethodContext, Redirect
from spyne.server import ServerBase
from spyne.protocol.http import HttpPattern
from spyne.const.http import gen_body_redirect, HTTP_301, HTTP_302, HTTP_303, \
    HTTP_307


class HttpRedirect(Redirect):
    def __init__(self, ctx, location, orig_exc=None, code=HTTP_302):
        super(HttpRedirect, self) \
                      .__init__(ctx, location, orig_exc=orig_exc)

        self.ctx = ctx
        self.location = location
        self.orig_exc = orig_exc
        self.code = code

    def do_redirect(self):
        if not isinstance(self.ctx.transport, HttpTransportContext):
            if self.orig_exc is not None:
                raise self.orig_exc
            raise TypeError(self.ctx.transport)

        self.ctx.transport.respond(self.code, location=self.location)

#
# Plagiarized HttpTransport.add_header() and _formatparam() function from
# Python 2.7 stdlib.
#
# Copyright (C) 2001-2007 Python Software Foundation
# Author: Barry Warsaw
# Contact: email-sig@python.org
#
def _formatparam(param, value=None, quote=True):
    """Convenience function to format and return a key=value pair.

    This will quote the value if needed or if quote is true.  If value is a
    three tuple (charset, language, value), it will be encoded according
    to RFC2231 rules.  If it contains non-ascii characters it will likewise
    be encoded according to RFC2231 rules, using the utf-8 charset and
    a null language.
    """
    if value is None or len(value) == 0:
        return param

    # A tuple is used for RFC 2231 encoded parameter values where items
    # are (charset, language, value).  charset is a string, not a Charset
    # instance.  RFC 2231 encoded values are never quoted, per RFC.
    if isinstance(value, tuple):
        # Encode as per RFC 2231
        param += '*'
        value = encode_rfc2231(value[2], value[0], value[1])
        return '%s=%s' % (param, value)

    try:
        value.encode('ascii')

    except UnicodeEncodeError:
        param += '*'
        value = encode_rfc2231(value, 'utf-8', '')
        return '%s=%s' % (param, value)

    # BAW: Please check this.  I think that if quote is set it should
    # force quoting even if not necessary.
    if quote or tspecials.search(value):
        return '%s="%s"' % (param, utils.quote(value))

    return '%s=%s' % (param, value)


class HttpTransportContext(TransportContext):
    """The abstract base class that is used in the transport attribute of the
    :class:`HttpMethodContext` class and its subclasses."""

    def __init__(self, parent, transport, request, content_type):
        super(HttpTransportContext, self).__init__(parent, transport, 'http')

        self.req = request
        """HTTP Request. This is transport-specific"""

        self.resp_headers = {}
        """HTTP Response headers."""

        self.mime_type = content_type

        self.resp_code = None
        """HTTP Response code."""

        self.wsdl = None
        """The WSDL document that is being returned. Only relevant when handling
        WSDL requests."""

        self.wsdl_error = None
        """The error when handling WSDL requests."""

    def get_mime_type(self):
        return self.resp_headers.get('Content-Type', None)

    def set_mime_type(self, what):
        self.resp_headers['Content-Type'] = what

    def respond(self, resp_code, **kwargs):
        self.resp_code = resp_code
        if resp_code in (HTTP_301, HTTP_302, HTTP_303, HTTP_307):
            l = kwargs.pop('location')
            self.resp_headers['Location'] = l
            self.parent.out_string = [gen_body_redirect(resp_code, l)]
            self.mime_type = 'text/html'

        else:
            # So that deserialization is skipped.
            self.parent.out_string = []

    def get_path(self):
        raise NotImplementedError()

    def get_request_method(self):
        raise NotImplementedError()

    def get_request_content_type(self):
        raise NotImplementedError()

    def get_path_and_qs(self):
        raise NotImplementedError()

    def get_cookie(self, key):
        raise NotImplementedError()

    def get_peer(self):
        raise NotImplementedError()

    @staticmethod
    def gen_header(_value, **kwargs):
        parts = []

        for k, v in kwargs.items():
            if v is None:
                parts.append(k.replace('_', '-'))

            else:
                parts.append(_formatparam(k.replace('_', '-'), v))

        if _value is not None:
            parts.insert(0, _value)

        return '; '.join(parts)

    def add_header(self, _name, _value, **kwargs):
        """Extended header setting.

        name is the header field to add.  keyword arguments can be used to set
        additional parameters for the header field, with underscores converted
        to dashes.  Normally the parameter will be added as key="value" unless
        value is None, in which case only the key will be added.  If a
        parameter value contains non-ASCII characters it can be specified as a
        three-tuple of (charset, language, value), in which case it will be
        encoded according to RFC2231 rules.  Otherwise it will be encoded using
        the utf-8 charset and a language of ''.

        Examples:

        msg.add_header('content-disposition', 'attachment', filename='bud.gif')
        msg.add_header('content-disposition', 'attachment',
                       filename=('utf-8', '', Fußballer.ppt'))
        msg.add_header('content-disposition', 'attachment',
                       filename='Fußballer.ppt'))
        """

        self.resp_headers[_name] = self.gen_header(_value, **kwargs)

    mime_type = property(
        lambda self: self.get_mime_type(),
        lambda self, what: self.set_mime_type(what),
    )
    """Provides an easy way to set outgoing mime type. Synonym for
    `content_type`"""

    content_type = mime_type
    """Provides an easy way to set outgoing mime type. Synonym for
    `mime_type`"""


class HttpMethodContext(MethodContext):
    """The Http-Specific method context. Http-Specific information is stored in
    the transport attribute using the :class:`HttpTransportContext` class.
    """

    # because ctor signatures differ between TransportContext and
    # HttpTransportContext, we needed a new variable
    TransportContext = None
    HttpTransportContext = HttpTransportContext

    def __init__(self, transport, req_env, content_type):
        super(HttpMethodContext, self).__init__(transport, MethodContext.SERVER)

        self.transport = self.HttpTransportContext(self, transport,
                                                          req_env, content_type)
        """Holds the HTTP-specific information"""

    def set_out_protocol(self, what):
        self._out_protocol = what
        if self._out_protocol.app is None:
            self._out_protocol.set_app(self.app)
        if isinstance(self.transport, HttpTransportContext):
            self.transport.set_mime_type(what.mime_type)

    out_protocol = property(MethodContext.get_out_protocol, set_out_protocol)
    """Assigning an out protocol overrides the mime type of the transport."""


class HttpBase(ServerBase):
    transport = 'http://schemas.xmlsoap.org/soap/http'

    def __init__(self, app, chunked=False,
                max_content_length=2 * 1024 * 1024,
                block_length=8 * 1024):
        super(HttpBase, self).__init__(app)

        self.chunked = chunked
        self.max_content_length = max_content_length
        self.block_length = block_length

        self._http_patterns = set()

        for k, v in self.app.interface.service_method_map.items():
            # p_ stands for primary
            p_method_descriptor = v[0]
            for patt in p_method_descriptor.patterns:
                if isinstance(patt, HttpPattern):
                    self._http_patterns.add(patt)

        # this makes sure similar addresses with patterns are evaluated after
        # addresses with wildcards, which puts the more specific addresses to
        # the front.
        self._http_patterns = list(reversed(sorted(self._http_patterns,
                                           key=lambda x: (x.address, x.host) )))

    def match_pattern(self, ctx, method='', path='', host=''):
        """Sets ctx.method_request_string if there's a match. It's O(n) which
        means you should keep your number of patterns as low as possible.

        :param ctx: A MethodContext instance
        :param method: The verb in the HTTP Request (GET, POST, etc.)
        :param host: The contents of the ``Host:`` header
        :param path: Path but not the arguments. (i.e. stuff before '?', if it's
            there)
        """

        if not path.startswith('/'):
            path = '/{}'.format(path)

        params = defaultdict(list)
        for patt in self._http_patterns:
            assert isinstance(patt, HttpPattern)

            if patt.verb is not None:
                match = patt.verb_re.match(method)
                if match is None:
                    continue
                if not (match.span() == (0, len(method))):
                    continue

                for k,v in match.groupdict().items():
                    params[k].append(v)

            if patt.host is not None:
                match = patt.host_re.match(host)
                if match is None:
                    continue
                if not (match.span() == (0, len(host))):
                    continue

                for k, v in match.groupdict().items():
                    params[k].append(v)

            if patt.address is None:
                if path.split('/')[-1] != patt.endpoint.name:
                    continue

            else:
                match = patt.address_re.match(path)
                if match is None:
                    continue

                if not (match.span() == (0, len(path))):
                    continue

                for k,v in match.groupdict().items():
                    params[k].append(v)

            d = patt.endpoint
            assert isinstance(d, MethodDescriptor)
            ctx.method_request_string = d.name

            break

        return params

    @property
    def has_patterns(self):
        return len(self._http_patterns) > 0




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#


"""The `spyne.util.xml` module contains various Xml and Xml Schema related
utility functions.
"""
from inspect import isgenerator

from lxml import etree

from os.path import dirname
from os.path import abspath

from spyne import ServiceBase, Application, srpc
from spyne.context import FakeContext
from spyne.interface import Interface
from spyne.interface.xml_schema import XmlSchema
from spyne.interface.xml_schema.parser import XmlSchemaParser, Thier_repr, PARSER
from spyne.protocol import ProtocolMixin
from spyne.protocol.cloth import XmlCloth

from spyne.protocol.xml import XmlDocument
from spyne.util.appreg import unregister_application
from spyne.util.six import BytesIO
from spyne.util.tlist import tlist


class FakeApplication(object):
    def __init__(self, default_namespace):
        self.tns = default_namespace
        self.services = ()
        self.classes = ()


def get_schema_documents(models, default_namespace=None):
    """Returns the schema documents in a dict whose keys are namespace prefixes
    and values are Element objects.

    :param models: A list of spyne.model classes that will be represented in
                   the schema.
    """

    if default_namespace is None:
        default_namespace = models[0].get_namespace()

    fake_app = FakeApplication(default_namespace)

    interface = Interface(fake_app)
    for m in models:
        m.resolve_namespace(m, default_namespace)
        interface.add_class(m)
    interface.populate_interface(fake_app)

    document = XmlSchema(interface)
    document.build_interface_document()

    return document.get_interface_document()


def get_validation_schema(models, default_namespace=None):
    """Returns the validation schema object for the given models.

    :param models: A list of spyne.model classes that will be represented in
                   the schema.
    """

    if default_namespace is None:
        default_namespace = models[0].get_namespace()

    fake_app = FakeApplication(default_namespace)

    interface = Interface(fake_app)
    for m in models:
        m.resolve_namespace(m, default_namespace)
        interface.add_class(m)

    schema = XmlSchema(interface)
    schema.build_validation_schema()

    return schema.validation_schema


def _dig(par):
    for elt in par:
        elt.tag = elt.tag.split('}')[-1]
        _dig(elt)


_xml_object = XmlDocument()


def get_object_as_xml(inst, cls=None, root_tag_name=None, no_namespace=False):
    """Returns an ElementTree representation of a
    :class:`spyne.model.complex.ComplexModel` subclass.

    :param inst: The instance of the class to be serialized.
    :param cls: The class to be serialized. Optional.
    :param root_tag_name: The root tag string to use. Defaults to the output of
        ``value.__class__.get_type_name_ns()``.
    :param no_namespace: When true, namespace information is discarded.
    """

    if cls is None:
        cls = inst.__class__

    parent = etree.Element("parent")
    _xml_object.to_parent(None, cls, inst, parent, cls.get_namespace(),
                                                                  root_tag_name)
    if no_namespace:
        _dig(parent)
        etree.cleanup_namespaces(parent)

    return parent[0]


def get_object_as_xml_polymorphic(inst, cls=None, root_tag_name=None,
                                                            no_namespace=False):
    """Returns an ElementTree representation of a
    :class:`spyne.model.complex.ComplexModel` subclass.

    :param inst: The instance of the class to be serialized.
    :param cls: The class to be serialized. Optional.
    :param root_tag_name: The root tag string to use. Defaults to the output of
        ``value.__class__.get_type_name_ns()``.
    :param no_namespace: When true, namespace information is discarded.
    """

    if cls is None:
        cls = inst.__class__

    if no_namespace:
        app = Application([ServiceBase], tns="",
                                     out_protocol=XmlDocument(polymorphic=True))
    else:
        tns = cls.get_namespace()
        if tns is None:
            raise ValueError(
                "Either set a namespace for %r or pass no_namespace=True"
                                                                      % (cls, ))

        class _DummyService(ServiceBase):
            @srpc(cls)
            def f(_):
                pass

        app = Application([_DummyService], tns=tns,
                                     out_protocol=XmlDocument(polymorphic=True))

    unregister_application(app)

    parent = etree.Element("parent", nsmap=app.interface.nsmap)

    app.out_protocol.to_parent(None, cls, inst, parent, cls.get_namespace(),
                                                                  root_tag_name)

    if no_namespace:
        _dig(parent)

    etree.cleanup_namespaces(parent)

    return parent[0]


def get_xml_as_object_polymorphic(elt, cls):
    """Returns a native :class:`spyne.model.complex.ComplexModel` child from an
    ElementTree representation of the same class.

    :param elt: The xml document to be deserialized.
    :param cls: The class the xml document represents.
    """

    tns = cls.get_namespace()
    if tns is None:
        raise ValueError("Please set a namespace for %r" % (cls, ))

    class _DummyService(ServiceBase):
        @srpc(cls)
        def f(_):
            pass

    app = Application([_DummyService], tns=tns,
                                      in_protocol=XmlDocument(polymorphic=True))

    unregister_application(app)

    return app.in_protocol.from_element(FakeContext(app=app), cls, elt)


def get_object_as_xml_cloth(inst, cls=None, no_namespace=False, encoding='utf8'):
    """Returns an ElementTree representation of a
    :class:`spyne.model.complex.ComplexModel` subclass.

    :param inst: The instance of the class to be serialized.
    :param cls: The class to be serialized. Optional.
    :param root_tag_name: The root tag string to use. Defaults to the output of
        ``value.__class__.get_type_name_ns()``.
    :param no_namespace: When true, namespace information is discarded.
    """

    if cls is None:
        cls = inst.__class__

    if cls.get_namespace() is None and no_namespace is None:
        no_namespace = True

    if no_namespace is None:
        no_namespace = False

    ostr = BytesIO()
    xml_cloth = XmlCloth(use_ns=(not no_namespace))
    ctx = FakeContext()
    with etree.xmlfile(ostr, encoding=encoding) as xf:
        ctx.outprot_ctx.doctype_written = False
        ctx.protocol.prot_stack = tlist([], ProtocolMixin)
        tn = cls.get_type_name()
        ret = xml_cloth.subserialize(ctx, cls, inst, xf, tn)

        assert not isgenerator(ret)

    return ostr.getvalue()


def get_xml_as_object(elt, cls):
    """Returns a native :class:`spyne.model.complex.ComplexModel` child from an
    ElementTree representation of the same class.

    :param elt: The xml document to be deserialized.
    :param cls: The class the xml document represents.
    """
    return _xml_object.from_element(None, cls, elt)


def parse_schema_string(s, files={}, repr_=Thier_repr(with_ns=False),
                                                         skip_errors=False):
    """Parses a schema string and returns a _Schema object.

    :param s: The string or bytes object that contains the schema document.
    :param files: A dict that maps namespaces to path to schema files that
        contain the schema document for those namespaces.
    :param repr_: A callable that functions as `repr`.
    :param skip_errors: Skip parsing errors and return a partial schema.
        See debug log for details.

    :return: :class:`spyne.interface.xml_schema.parser._Schema` instance.
    """

    elt = etree.fromstring(s, parser=PARSER)
    return XmlSchemaParser(files, repr_=repr_,
                            skip_errors=skip_errors).parse_schema(elt)


def parse_schema_element(elt, files={}, repr_=Thier_repr(with_ns=False),
                                                         skip_errors=False):
    """Parses a `<xs:schema>` element and returns a _Schema object.

    :param elt: The `<xs:schema>` element, an lxml.etree._Element instance.
    :param files: A dict that maps namespaces to path to schema files that
        contain the schema document for those namespaces.
    :param repr_: A callable that functions as `repr`.
    :param skip_errors: Skip parsing errors and return a partial schema.
        See debug log for details.

    :return: :class:`spyne.interface.xml_schema.parser._Schema` instance.
    """

    return XmlSchemaParser(files, repr_=repr_,
                            skip_errors=skip_errors).parse_schema(elt)


def parse_schema_file(file_name, files=None, repr_=Thier_repr(with_ns=False),
                                                         skip_errors=False):
    """Parses a schema file and returns a _Schema object. Schema files typically
    have the `*.xsd` extension.

    :param file_name: The path to the file that contains the schema document
        to be parsed.
    :param files: A dict that maps namespaces to path to schema files that
        contain the schema document for those namespaces.
    :param repr_: A callable that functions as `repr`.
    :param skip_errors: Skip parsing errors and return a partial schema.
        See debug log for details.

    :return: :class:`spyne.interface.xml_schema.parser._Schema` instance.
    """

    if files is None:
        files = dict()

    elt = etree.fromstring(open(file_name, 'rb').read(), parser=PARSER)
    wd = abspath(dirname(file_name))
    return XmlSchemaParser(files, wd, repr_=repr_,
                            skip_errors=skip_errors).parse_schema(elt)
